/*
 * Copyright (C) 2006 Red Hat, Inc.
 *
 * This file is released under the LGPL.
 */

#include <linux/blkdev.h> /* for sector_div, which is used in dm-dirty-log.h */
#include <linux/bio.h>
#include <linux/dm-dirty-log.h>
#include <linux/device-mapper.h>

#include "dm-clog-tfr.h"

struct flush_entry {
	int type;
	region_t region;
	struct list_head list;
};

struct log_c {
	struct list_head list;

	struct dm_target *ti;
	uint32_t region_size;
	region_t region_count;
	char uuid[DM_UUID_LEN];
	uint32_t uuid_instance;

	char *ctr_str; /* Gives ability to restart if userspace dies */
	uint32_t ctr_size;

	/*
	 * in_sync_hint gets set when doing is_remote_recovering.  It
	 * represents the first region that needs recovery.  IOW, the
	 * first zero bit of sync_bits.  This can be useful for to limit
	 * traffic for calls like is_remote_recovering and get_resync_work,
	 * but be take care in its use for anything else.
	 */
	uint64_t in_sync_hint;

	spinlock_t flush_lock;
	struct list_head flush_list;  /* only for clear and mark requests */

	struct dm_dev *disk_log;
};

static struct list_head log_list_head;
static spinlock_t log_list_lock;
static mempool_t *flush_entry_pool = NULL;

static void *flush_entry_alloc(gfp_t gfp_mask, void *pool_data)
{
	return kmalloc(sizeof(struct flush_entry), gfp_mask);
}

static void flush_entry_free(void *element, void *pool_data)
{
	kfree(element);
}

int cluster_do_request(struct log_c *lc, const char *uuid, int request_type,
		       char *data, int data_size, char *rdata, int *rdata_size)
{
	int r;

	/*
	 * If the server isn't there, -ESRCH is returned,
	 * and we must keep trying until the server is
	 * restored.
	 */
retry:
	r = dm_clog_consult_server(uuid, lc->uuid_instance, request_type, data,
				   data_size, rdata, rdata_size);

	if (r != -ESRCH)
		return r;

	DMERR(" Userspace cluster log server not found.");
	while (1) {
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(2*HZ);
		DMWARN("Attempting to contact cluster log server...");
		r = dm_clog_consult_server(uuid, lc->uuid_instance,
					   DM_CLOG_CTR, lc->ctr_str,
					   lc->ctr_size, NULL, NULL);
		if (!r)
			break;
	}
	DMINFO("Reconnected to cluster log server... CTR complete");
	r = dm_clog_consult_server(uuid, lc->uuid_instance,
				   DM_CLOG_RESUME, NULL,
				   0, NULL, NULL);
	if (!r)
		goto retry;

	DMERR("Error trying to resume cluster log: %d", r);

	return -ESRCH;
}

static int cluster_ctr(struct dm_dirty_log *log, struct dm_target *ti,
		       unsigned int argc, char **argv,
		       struct dm_dev *disk_log)
{
	int i;
	int r = 0;
	int str_size;
	int offset = (disk_log) ? 1 : 0;
	char *ctr_str = NULL;
	struct log_c *lc = NULL, *tmp;
	uint32_t region_size;
	region_t region_count;

	/* Already checked argument count */

	if (sscanf(argv[offset], "%u", &region_size) != 1) {
		DMWARN("Invalid region size string");
		return -EINVAL;
	}

	region_count = dm_sector_div_up(ti->len, region_size);

	lc = kmalloc(sizeof(*lc), GFP_KERNEL);
	if (!lc) {
		DMWARN("Unable to allocate cluster log context.");
		return -ENOMEM;
	}

	lc->ti = ti;
	lc->region_size = region_size;
	lc->region_count = region_count;
	lc->disk_log = disk_log;

	/* FIXME: Need to check size of uuid arg */
	memcpy(lc->uuid, argv[1 + offset], DM_UUID_LEN);
	spin_lock_init(&lc->flush_lock);
	INIT_LIST_HEAD(&lc->flush_list);

	lc->uuid_instance = 1;
	spin_lock(&log_list_lock);
	list_for_each_entry(tmp, &log_list_head, list)
		if (!strncmp(tmp->uuid, lc->uuid, DM_UUID_LEN) &&
		    (tmp->uuid_instance >= lc->uuid_instance))
			lc->uuid_instance = tmp->uuid_instance + 1;
	list_add(&lc->list, &log_list_head);
	spin_unlock(&log_list_lock);

	for (i = 0, str_size = 0; i < argc; i++)
		str_size += strlen(argv[i]) + 1; /* +1 for space between args */

	str_size += 20; /* Max number of chars in a printed u64 number */

	ctr_str = kzalloc(str_size, GFP_KERNEL);
	if (!ctr_str) {
		DMWARN("Unable to allocate memory for constructor string");
		kfree(lc);
		return -ENOMEM;
	}

	for (i = 0, str_size = 0; i < argc; i++)
		str_size += sprintf(ctr_str + str_size, "%s ", argv[i]);
	str_size += sprintf(ctr_str + str_size, "%llu", ti->len);

	/* Send table string */
	r = dm_clog_consult_server(lc->uuid, lc->uuid_instance, DM_CLOG_CTR,
				   ctr_str, str_size, NULL, NULL);

	if (r == -ESRCH)
		DMERR(" Userspace cluster log server not found");

	if (r) {
		spin_lock(&log_list_lock);
		list_del(&lc->list);
		spin_unlock(&log_list_lock);
		kfree(lc);
		kfree(ctr_str);
	} else {
		lc->ctr_str = ctr_str;
		lc->ctr_size = str_size;
		log->context = lc;
	}

	return r;
}

/*
 * cluster_core_ctr
 * @log
 * @ti
 * @argc
 * @argv
 *
 * argv contains:
 *   <region_size> <uuid> [[no]sync] [block_on_error]
 *
 * Returns: 0 on success, -XXX on failure
 */
static int cluster_core_ctr(struct dm_dirty_log *log, struct dm_target *ti,
			    unsigned int argc, char **argv)
{
	int i, r;
	if ((argc < 2) || (argc > 4)) {
		DMERR("Too %s arguments to clustered-core mirror log type.",
		      (argc < 2) ? "few" : "many");
		DMERR("  %d arguments supplied:", argc);
		for (i = 0; i < argc; i++)
			DMERR("    %s", argv[i]);
		return -EINVAL;
	}

	r = cluster_ctr(log, ti, argc, argv, NULL);

	return r;
}


/*
 * cluster_core_ctr
 * @log
 * @ti
 * @argc
 * @argv
 *
 * argv contains:
 *   <disk> <region_size> <uuid> [[no]sync] [block_on_error]
 *
 * Returns: 0 on success, -XXX on failure
 */
static int cluster_disk_ctr(struct dm_dirty_log *log, struct dm_target *ti,
			    unsigned int argc, char **argv)
{
	int r, i;
	struct dm_dev *dev;

	if ((argc < 3) || (argc > 5)) {
		DMERR("Too %s arguments to clustered-disk mirror log type.",
		      (argc < 3) ? "few" : "many");
		DMERR("  %d arguments supplied:", argc);
		for (i = 0; i < argc; i++)
			DMERR("    %s", argv[i]);
		return -EINVAL;
	}

	r = dm_get_device(ti, argv[0], 0, 0, FMODE_READ | FMODE_WRITE, &dev);
	if (r)
		return r;

	r = cluster_ctr(log, ti, argc, argv, dev);
	if (r)
		dm_put_device(ti, dev);

	return r;
}

/*
 * cluster_dtr
 * @log
 */
static void cluster_dtr(struct dm_dirty_log *log)
{
	int r;
	struct log_c *lc = (struct log_c *)log->context;

	r = dm_clog_consult_server(lc->uuid, lc->uuid_instance, DM_CLOG_DTR,
				   NULL, 0,
				   NULL, NULL);

	/* FIXME: What do we do on failure? */

	spin_lock(&log_list_lock);
	list_del(&lc->list);
	spin_unlock(&log_list_lock);
	if (lc->disk_log)
		dm_put_device(lc->ti, lc->disk_log);
	kfree(lc->ctr_str);
	kfree(lc);

	return;
}

/*
 * cluster_presuspend
 * @log
 */
static int cluster_presuspend(struct dm_dirty_log *log)
{
	int r;
	struct log_c *lc = (struct log_c *)log->context;

	r = dm_clog_consult_server(lc->uuid, lc->uuid_instance,
				   DM_CLOG_PRESUSPEND,
				   NULL, 0,
				   NULL, NULL);

	return r;
}

/*
 * cluster_postsuspend
 * @log
 */
static int cluster_postsuspend(struct dm_dirty_log *log)
{
	int r;
	struct log_c *lc = (struct log_c *)log->context;

	r = dm_clog_consult_server(lc->uuid, lc->uuid_instance,
				   DM_CLOG_POSTSUSPEND,
				   NULL, 0,
				   NULL, NULL);

	return r;
}

/*
 * cluster_resume
 * @log
 */
static int cluster_resume(struct dm_dirty_log *log)
{
	int r;
	struct log_c *lc = (struct log_c *)log->context;

	lc->in_sync_hint = 0;
	r = dm_clog_consult_server(lc->uuid, lc->uuid_instance, DM_CLOG_RESUME,
				   NULL, 0,
				   NULL, NULL);

	return r;
}

/*
 * cluster_get_region_size
 * @log
 *
 * Only called during mirror construction, ok to block.
 *
 * Returns: region size (doesn't fail)
 */
static uint32_t cluster_get_region_size(struct dm_dirty_log *log)
{
	struct log_c *lc = (struct log_c *)log->context;

	return lc->region_size;
}

/*
 * cluster_is_clean
 * @log
 * @region
 *
 * Check whether a region is clean.  If there is any sort of
 * failure when consulting the server, we return not clean.
 *
 * Returns: 1 if clean, 0 otherwise
 */
static int cluster_is_clean(struct dm_dirty_log *log, region_t region)
{
	int r;
	int is_clean;
	int rdata_size;
	struct log_c *lc = (struct log_c *)log->context;

	rdata_size = sizeof(is_clean);
	r = cluster_do_request(lc, lc->uuid, DM_CLOG_IS_CLEAN,
			       (char *)&region, sizeof(region),
			       (char *)&is_clean, &rdata_size);

	return (r) ? 0 : is_clean;
}

/*
 * cluster_in_sync
 * @log
 * @region
 * @can_block: if set, return immediately
 *
 * Check if the region is in-sync.  If there is any sort
 * of failure when consulting the server, we assume that
 * the region is not in sync.
 *
 * Returns: 1 if in-sync, 0 if not-in-sync, -EWOULDBLOCK
 */
static int cluster_in_sync(struct dm_dirty_log *log, region_t region, int can_block)
{
	int r;
	int in_sync;
	int rdata_size;
	struct log_c *lc = (struct log_c *)log->context;

	/*
	 * We can never respond directly - even if in_sync_hint is
	 * set.  This is because another machine could see a device
	 * failure and mark the region out-of-sync.  If we don't go
	 * to userspace to ask, we might think the region is in-sync
	 * and allow a read to pick up data that is stale.  (This is
	 * very unlikely if a device actually fails; but it is very
	 * likely if a connection to one device from one machine fails.)
	 *
	 * There still might be a problem if the mirror caches the region
	 * state as in-sync... but then this call would not be made.  So,
	 * that is a mirror problem.
	 */
	if (!can_block)
		return -EWOULDBLOCK;

	rdata_size = sizeof(in_sync);
	r = cluster_do_request(lc, lc->uuid, DM_CLOG_IN_SYNC,
			       (char *)&region, sizeof(region),
			       (char *)&in_sync, &rdata_size);
	return (r) ? 0 : in_sync;
}

/*
 * cluster_flush
 * @log
 *
 * This function is ok to block.
 * The flush happens in two stages.  First, it sends all
 * clear/mark requests that are on the list.  Then it
 * tells the server to commit them.  This gives the
 * server a chance to optimise the commit to the cluster
 * and/or disk, instead of doing it for every request.
 *
 * Additionally, we could implement another thread that
 * sends the requests up to the server - reducing the
 * load on flush.  Then the flush would have less in
 * the list and be responsible for the finishing commit.
 *
 * Returns: 0 on success, < 0 on failure
 */
static int cluster_flush(struct dm_dirty_log *log)
{
	int r = 0;
	int flags;
	struct log_c *lc = (struct log_c *)log->context;
	LIST_HEAD(flush_list);
	struct flush_entry *fe, *tmp_fe;

	spin_lock_irqsave(&lc->flush_lock, flags);
	list_splice_init(&lc->flush_list, &flush_list);
	spin_unlock_irqrestore(&lc->flush_lock, flags);

	if (list_empty(&flush_list))
		return 0;

	/*
	 * FIXME: Count up requests, group request types,
	 * allocate memory to stick all requests in and
	 * send to server in one go.  Failing the allocation,
	 * do it one by one.
	 */

	list_for_each_entry(fe, &flush_list, list) {
		r = cluster_do_request(lc, lc->uuid, fe->type,
				       (char *)&fe->region,
				       sizeof(fe->region),
				       NULL, NULL);
		if (r)
			goto fail;
	}

	r = cluster_do_request(lc, lc->uuid, DM_CLOG_FLUSH,
			       NULL, 0, NULL, NULL);

fail:
	/*
	 * We can safely remove these entries, even if failure.
	 * Calling code will receive an error and will know that
	 * the log facility has failed.
	 */
	list_for_each_entry_safe(fe, tmp_fe, &flush_list, list) {
		list_del(&fe->list);
		mempool_free(fe, flush_entry_pool);
	}

	if (r)
		dm_table_event(lc->ti->table);

	return r;
}

/*
 * cluster_mark_region
 * @log
 * @region
 *
 * This function should avoid blocking unless absolutely required.
 * (Memory allocation is valid for blocking.)
 */
static void cluster_mark_region(struct dm_dirty_log *log, region_t region)
{
	int flags;
	struct log_c *lc = (struct log_c *)log->context;
	struct flush_entry *fe;

	/* Wait for an allocation, but _never_ fail */
	fe = mempool_alloc(flush_entry_pool, GFP_NOIO);
	BUG_ON(!fe);

	spin_lock_irqsave(&lc->flush_lock, flags);
	fe->type = DM_CLOG_MARK_REGION;
	fe->region = region;
	list_add(&fe->list, &lc->flush_list);
	spin_unlock_irqrestore(&lc->flush_lock, flags);

	return;
}

/*
 * cluster_clear_region
 * @log
 * @region
 *
 * This function must not block.
 * So, the alloc can't block.  In the worst case, it is ok to
 * fail.  It would simply mean we can't clear the region.
 * Does nothing to current sync context, but does mean
 * the region will be re-sync'ed on a reload of the mirror
 * even though it is in-sync.
 */
static void cluster_clear_region(struct dm_dirty_log *log, region_t region)
{
	int flags;
	struct log_c *lc = (struct log_c *)log->context;
	struct flush_entry *fe;

	/*
	 * If we fail to allocate, we skip the clearing of
	 * the region.  This doesn't hurt us in any way, except
	 * to cause the region to be resync'ed when the
	 * device is activated next time.
	 */
	fe = mempool_alloc(flush_entry_pool, GFP_ATOMIC);
	if (!fe) {
		DMERR("Failed to allocate memory to clear region.");
		return;
	}

	spin_lock_irqsave(&lc->flush_lock, flags);
	fe->type = DM_CLOG_CLEAR_REGION;
	fe->region = region;
	list_add(&fe->list, &lc->flush_list);
	spin_unlock_irqrestore(&lc->flush_lock, flags);

	return;
}

/*
 * cluster_get_resync_work
 * @log
 * @region
 *
 * Get a region that needs recovery.  It is valid to return
 * an error for this function.
 *
 * Returns: 1 if region filled, 0 if no work, <0 on error
 */
static int cluster_get_resync_work(struct dm_dirty_log *log, region_t *region)
{
	int r;
	int rdata_size;
	struct log_c *lc = (struct log_c *)log->context;
	struct { 
		int32_t i;
		uint32_t arch_padding;
		region_t r;
	} pkg;

	if (lc->in_sync_hint >= lc->region_count)
		return 0;

	rdata_size = sizeof(pkg);
	r = cluster_do_request(lc, lc->uuid, DM_CLOG_GET_RESYNC_WORK,
			       NULL, 0,
			       (char *)&pkg, &rdata_size);

	*region = pkg.r;
	return (r) ? r : pkg.i;
}

/*
 * cluster_set_region_sync
 * @log
 * @region
 * @in_sync
 *
 * Set the sync status of a given region.  This function
 * must not fail.
 */
static void cluster_set_region_sync(struct dm_dirty_log *log,
				    region_t region, int in_sync)
{
	int r;
	struct log_c *lc = (struct log_c *)log->context;
	struct {
		region_t r;
		uint32_t arch_padding;
		int32_t i;
	} pkg;

	pkg.r = region;
	pkg.i = in_sync;

	r = cluster_do_request(lc, lc->uuid, DM_CLOG_SET_REGION_SYNC,
			       (char *)&pkg, sizeof(pkg),
			       NULL, NULL);

	/* FIXME: It would be nice to be able to report failures */
	return;
}

/*
 * cluster_get_sync_count
 * @log
 *
 * If there is any sort of failure when consulting the server,
 * we assume that the sync count is zero.
 *
 * Returns: sync count on success, 0 on failure
 */
static region_t cluster_get_sync_count(struct dm_dirty_log *log)
{
	int r;
	int rdata_size;
	region_t sync_count;
	struct log_c *lc = (struct log_c *)log->context;

	rdata_size = sizeof(sync_count);
	r = cluster_do_request(lc, lc->uuid, DM_CLOG_GET_SYNC_COUNT,
			       NULL, 0,
			       (char *)&sync_count, &rdata_size);

	if (r)
		return 0;

	if (sync_count >= lc->region_count)
		lc->in_sync_hint = lc->region_count;
	/*
	 * get_sync_count is never called after the
	 * initial sync=1
	else
		lc->in_sync_hint = 0;
	*/

	return sync_count;
}

/*
 * cluster_status
 * @log
 * @status_type
 * @result
 * @maxlen
 *
 * Returns: amount of space consumed
 */
static int cluster_status(struct dm_dirty_log *log, status_type_t status_type,
			  char *result, unsigned int maxlen)
{
	int r = 0;
	unsigned int sz = maxlen;
	struct log_c *lc = (struct log_c *)log->context;

	switch(status_type) {
	case STATUSTYPE_INFO:
		r = cluster_do_request(lc, lc->uuid, DM_CLOG_STATUS_INFO,
				       NULL, 0,
				       result, &sz);
		/*
		 * FIXME: If we fail to contact server, we should still
		 * populate this with parsible results
		 */
		break;
	case STATUSTYPE_TABLE:
		r = cluster_do_request(lc, lc->uuid, DM_CLOG_STATUS_TABLE,
				       NULL, 0,
				       result, &sz);
		break;
	}
	return (r) ? 0: sz;
}

static int cluster_get_failure_response(struct dm_dirty_log *log)
{
	/* This type of return is not an option */
	return DMLOG_IOERR_BLOCK;
}

static int cluster_is_remote_recovering(struct dm_dirty_log *log, region_t region)
{
	int r;
	struct log_c *lc = (struct log_c *)log->context;
	static unsigned long long limit = 0;
	struct {
		int32_t is_recovering;
		uint32_t arch_padding;
		uint64_t in_sync_hint;
	} pkg;
	int rdata_size = sizeof(pkg);

	/*
	 * Once the mirror has been reported to be in-sync,
	 * it will never again ask for recovery work.  So,
	 * we can safely say there is not a remote machine
	 * recovering if the device is in-sync.  (in_sync_hint
	 * must be reset at resume time.)
	 */
	if (region < lc->in_sync_hint)
		return 0;
	else if (jiffies < limit)
		return 1;

	limit = jiffies + (HZ / 4);
	r = cluster_do_request(lc, lc->uuid, DM_CLOG_IS_REMOTE_RECOVERING,
			       (char *)&region, sizeof(region),
			       (char *)&pkg, &rdata_size);
	if (r)
		return 1;

	lc->in_sync_hint = pkg.in_sync_hint;

	return pkg.is_recovering;
}

static struct dm_dirty_log_type _clustered_core_type = {
	.name = "clustered-core",
	.module = THIS_MODULE,
	.ctr = cluster_core_ctr,
	.dtr = cluster_dtr,
	.presuspend = cluster_presuspend,
	.postsuspend = cluster_postsuspend,
	.resume = cluster_resume,
	.get_region_size = cluster_get_region_size,
	.is_clean = cluster_is_clean,
	.in_sync = cluster_in_sync,
	.flush = cluster_flush,
	.mark_region = cluster_mark_region,
	.clear_region = cluster_clear_region,
	.get_resync_work = cluster_get_resync_work,
	.set_region_sync = cluster_set_region_sync,
	.get_sync_count = cluster_get_sync_count,
	.status = cluster_status,
	.get_failure_response = cluster_get_failure_response,
	.is_remote_recovering = cluster_is_remote_recovering,
};

static struct dm_dirty_log_type _clustered_disk_type = {
	.name = "clustered-disk",
	.module = THIS_MODULE,
	.ctr = cluster_disk_ctr,
	.dtr = cluster_dtr,
	.presuspend = cluster_presuspend,
	.postsuspend = cluster_postsuspend,
	.resume = cluster_resume,
	.get_region_size = cluster_get_region_size,
	.is_clean = cluster_is_clean,
	.in_sync = cluster_in_sync,
	.flush = cluster_flush,
	.mark_region = cluster_mark_region,
	.clear_region = cluster_clear_region,
	.get_resync_work = cluster_get_resync_work,
	.set_region_sync = cluster_set_region_sync,
	.get_sync_count = cluster_get_sync_count,
	.status = cluster_status,
	.get_failure_response = cluster_get_failure_response,
	.is_remote_recovering = cluster_is_remote_recovering,
};

static int __init cluster_dirty_log_init(void)
{
	int r = 0;

	INIT_LIST_HEAD(&log_list_head);
	spin_lock_init(&log_list_lock);

	flush_entry_pool = mempool_create(100, flush_entry_alloc,
					  flush_entry_free, NULL);

	if (!flush_entry_pool) {
		DMWARN("Unable to create flush_entry_pool:  No memory.");
		return -ENOMEM;
	}

	r = dm_clog_tfr_init();
	if (r) {
		DMWARN("Unable to initialize cluster log communications");
		mempool_destroy(flush_entry_pool);
		return r;
	}

	r = dm_dirty_log_type_register(&_clustered_core_type);
	if (r) {
		DMWARN("Couldn't register clustered-core dirty log type");
		dm_clog_tfr_exit();
		mempool_destroy(flush_entry_pool);
		return r;
	}

	r = dm_dirty_log_type_register(&_clustered_disk_type);
	if (r) {
		DMWARN("Couldn't register clustered-disk dirty log type");
		dm_dirty_log_type_unregister(&_clustered_core_type);
		dm_clog_tfr_exit();
		mempool_destroy(flush_entry_pool);
		return r;
	}

	DMINFO("(built %s %s) installed", __DATE__, __TIME__);
	return 0;
}

static void __exit cluster_dirty_log_exit(void)
{
	dm_dirty_log_type_unregister(&_clustered_disk_type);
	dm_dirty_log_type_unregister(&_clustered_core_type);
	dm_clog_tfr_exit();
	mempool_destroy(flush_entry_pool);
	DMINFO("(built %s %s) removed", __DATE__, __TIME__);
	return;
}

module_init(cluster_dirty_log_init);
module_exit(cluster_dirty_log_exit);

MODULE_DESCRIPTION(DM_NAME " mirror cluster-aware log");
MODULE_AUTHOR("Jonathan Brassow");
MODULE_LICENSE("GPL");
